//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2011 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Frequency offset introduced 26/05/2011.
// Convolutional coding introduced 6/12/2011.
// Preamble amplitude adjustment (factor of 0.5) done on 02/04/2012.
// This code uses genlib() which generates binary files and is hence faster
// than dmt10. All if-else statements have been removed. ML channel estimate
// and non-ideal timing is assumed.
// The demodulation-matched filtering operation does not result in aliasing.
// Normalize_Alpha_Beta code incorporated in Dec1_Process and Dec2_Process.
// Postamble introduced on 01/08/2012.
// Robust turbo coding (normalizing exponents) introduced on 12/12/2012.
// Overall maximum is used to normalize the exponents.
// Rate increased to 1 on 19/02/2013.
// Correction to normalization of the exponents done on 29/05/2013. Collect
// the maximum exponent from each column.
// This is identical to dmt17_e, except that different types of turbo codes
// can be used. "ldiv" is not used for generating the impulse response of the
// constituent convolutional codes. Introduced on 5th Dec 2013.
// Code tested on SCILAB ver 5.3.3.
//******************************************************************************
//clear;
//getd(pwd());                    // Get all functions in the current directory.
timer();
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=100;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                     // Length of the preamble.
OFDM_SIZE=4096;           // Length of OFDM symbol, excluding the cyclic prefix.
POST_AMB_SIZE=256;                         // Length of postamble, must be even.
BUF_SIZE=4;                 // Length of buffer before and after the data frame.
DATA_SIZE=OFDM_SIZE-POST_AMB_SIZE-2*BUF_SIZE;   // Length of data, must be even.
PD_SIZE=DATA_SIZE+POST_AMB_SIZE;               // Length of postamble plus data.
PRE_AMB_QPSK_AMP=sqrt(PRE_AMB_SIZE/OFDM_SIZE);     // Amp. of preamble QPSK sym.
CHAN_MEM=9;                                       // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=OFDM_SIZE+PRE_AMB_SIZE+EST_CHAN_MEM;                   // Frame size.
IP_FAC=16;                                    // Interpolation factor at the rx.
IP_FAC_ZP=IP_FAC-1;                                   // Zero padding at the rx.
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF=0.04;           // Freq. offset lies in the interval [-0.04, 0.04] radians.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_LEN-1; // Size of rx. vec. for freq. est.
MAX_FOFF=0.04;                                            // Max . freq. offset.
MAX_FOFF2=0.005;                               // Max. freq. offset in 2nd step.
ML_FOFF1_BINS=64;                   // No. of freq. bins in 1st step of ML est.
ML_FOFF2_BINS=64;                   // No. of freq. bins in 2nd step of ML est.
//******************************************************************************
// SNR parameters.
//******************************************************************************
SNR_DB=6.0;                                                       // SNR in dB.
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));    // Var per dimension of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
for i=1:CONSTELL_SIZE
 QPSK_SYM(i,:)=QPSK_Map(i)*ones(1,DATA_SIZE);
end
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
[Re_Pre_Amb_Sym,Im_Pre_Amb_Sym,Pre_Amb_Sym]=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym=Pre_Amb_Sym*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1).'; // IFFT. There is division by PRE_AMB_SIZE.
//******************************************************************************
// Generate postamble QPSK symbols.
//******************************************************************************
[Re_Post_Amb_Sym,Im_Post_Amb_Sym,Post_Amb_Sym]=Get_QPSK_Sym(POST_AMB_SIZE);
//******************************************************************************
// Generate buffer QPSK symbols.
//******************************************************************************
[Re_Buf_Sym1,Im_Buf_Sym1,Buf_Sym1]=Get_QPSK_Sym(BUF_SIZE);      // Start buffer.
[Re_Buf_Sym2,Im_Buf_Sym2,Buf_Sym2]=Get_QPSK_Sym(BUF_SIZE);        // End buffer.
//******************************************************************************
// Get channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE-EST_CHAN_MEM,EST_CHAN_LEN);     // Initialize.
for i=1:PRE_AMB_SIZE-EST_CHAN_MEM
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT(EST_CHAN_LEN+i-1:-1:i);
end
//******************************************************************************
// Least squares channel estimation matrix.
//******************************************************************************
 Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// Get the matched filter.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
MF_Coeff=MF_Coeff/MF_Energy;                       // Normalized to unit energy.
RX_MF_SIZE=FOFF_EST_RX_SIZE;                                  // Size of rx. MF.
//******************************************************************************
// Get the frequency offset matrix for a fixed frequency offset.
//******************************************************************************
//FOFF_Vec=0:FOFF_VEC_SIZE-1;
//FOFF_Vec=FOFF_Vec*%i*FOFF;
//FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x;
clear y;
//******************************************************************************
// Frequency offset estimation matrix for step 2.
//******************************************************************************
Freq_Step2=2*MAX_FOFF2/ML_FOFF2_BINS;
y2=[0:ML_FOFF2_BINS]';
y2=y2*Freq_Step2-MAX_FOFF2;     // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix2=y2*x;
FOFF_Est_Matrix2=exp(-%i*FOFF_Est_Matrix2);
clear x;
clear y2;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
RMS_Err_FOFF1=0.0;         // RMS frequency offset estimation error in 1st step.
Max_Err_FOFF1=-1.0;    // Maximum frequency offset estimation error in 1st step.
RMS_Err_FOFF2=0.0;         // RMS frequency offset estimation error in 2nd step.
Max_Err_FOFF2=-1.0;    // Maximum frequency offset estimation error in 2nd step.
RMS_Err_FOFF3=0.0;         // RMS freq off est error in 2nd step, coherent case.
Max_Err_FOFF3=-1.0;    // Maximum freq off est error in 2nd step, coherent case.
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D^3+D^4)/(1+D+D^3+D^4)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 1 1 1 0 1];                                            // 1st term is D^0.
DC=[1 1 1 1 1 1];                                            // 1st term is D^0.
//NP=poly(NC,'D','coeff');                              // Numerator polynomial.
//DP=poly(DC,'D','coeff');                            // Denominator polynomial.
//RSC_Gen_Vec=modulo(abs(ldiv(NP,DP,DATA_SIZE)),2);   // Column vector of coeff.
NC_Size=size(NC);
NUM_INPUTS=2;                                      // No. of input combinations.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op,Ga_Inx,Gb_Inx]=Get_Trellis();
RSC_Gen_Vec=Get_RSC_Gen_Vec();                          // Get impulse response.
pause;
//******************************************************************************
// Get the interleaver and deinterleaver map for the turbo code.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(DATA_SIZE);
//******************************************************************************
// Get the interleaver and deinterleaver map for OFDM.
//******************************************************************************
[OFDM_Iv_Map,OFDM_Div_Map]=Get_OFDM_Iv_Div_Map(PD_SIZE);
//******************************************************************************
// Cyclic prefix (CP) indices.
//******************************************************************************
CP_ST_INX=OFDM_SIZE-EST_CHAN_MEM+1;         // Start index for extraction of CP.
CP_ED_INX=OFDM_SIZE;                          // End index for extraction of CP.
//******************************************************************************
// Starting and ending indices for the postamble in the data frame.
//******************************************************************************
POST_START=DATA_SIZE+1;
POST_END=PD_SIZE;
//******************************************************************************
// Interleaved indices of the postamble.
//******************************************************************************
Div_Post_Inx=OFDM_Div_Map(POST_START:POST_END);
//******************************************************************************
// Resizing of the deinterleaved FFT output.
//******************************************************************************
//Div_Rx_FFT_Op=zeros(IP_FAC,2*OFDM_SIZE);
//******************************************************************************
// Starting and ending indices for the data (consisting of information symbols
// and interleaved information symbols).
//******************************************************************************
//DATA_START=BUF_SIZE+1;
//DATA_END=BUF_SIZE+2*DATA_SIZE;
//******************************************************************************
// Starting and ending indices for information symbols in the data frame.
//******************************************************************************
//INFO_START=BUF_SIZE+1;
//INFO_END=INFO_START+DATA_SIZE-1;
//******************************************************************************
// Starting and ending indices for interleaved information symbols in the
// data frame excluding the buffer and postamble.
//******************************************************************************
GINX1=(DATA_SIZE/2)+1;
GINX2=DATA_SIZE;
//******************************************************************************
// Minimum exponent.
//******************************************************************************
MIN_EXPO= -30.0;
//******************************************************************************
// Initialize gammas.
//******************************************************************************
My_Gamma1=zeros(CONSTELL_SIZE,DATA_SIZE);
My_Gamma2=zeros(CONSTELL_SIZE,DATA_SIZE);
//******************************************************************************
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Erasure=0;                                          // Number of frame erasures.
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
//******************************************************************************
// Get the random frequency offset and the frequency offset matrix.
//******************************************************************************
 FOFF=2.0*MAX_FOFF*(grand(1,1,'def')-0.5);
// FOFF=0.0;
 FOFF_Vec=0:FOFF_VEC_SIZE-1;
 FOFF_Vec=FOFF_Vec*%i*FOFF;
 FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Initialize estimates of the frequency offset.
//******************************************************************************
 Est_FOFF1=0.0;
 Est_FOFF2=0.0;
//******************************************************************************
// Generate coded QPSK symbols and append postamble.
//******************************************************************************
 [Coded_Tx_Sym,My_Data]=Get_Coded_QPSK(DATA_SIZE,RSC_Gen_Vec,Div_Map);
 Coded_Tx_Sym=[Coded_Tx_Sym Post_Amb_Sym];
//******************************************************************************
// Interleaving and add buffer symbols.
//******************************************************************************
 Tx_Sym=[Buf_Sym1 Coded_Tx_Sym(OFDM_Iv_Map(:)) Buf_Sym2];
//******************************************************************************
 Sym_IFFT=fft(Tx_Sym.',1).';            // IFFT. There is division by OFDM_SIZE.
//******************************************************************************
// Add cyclic prefix and preamble.
//******************************************************************************
 Chan_Ip=[Pre_Amb_IFFT Sym_IFFT(CP_ST_INX:CP_ED_INX) Sym_IFFT];
//******************************************************************************
// Get channel.
//******************************************************************************
 [Re_Chan,Im_Chan,Chan]=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan zeros(1,OFDM_SIZE-CHAN_LEN)];             // Zero padding.
 Chan_FFT=fft(Chan_Ext_Zeros.',-1).';                    // OFDM_SIZE-point FFT.
//******************************************************************************
 Chan_Op=Get_Channel_Op(Chan,Chan_Ip,FOFF_Vec,NOISE_MEAN,NOISE_STD);
 FOFF_Est_Rx=Chan_Op(1:FOFF_EST_RX_SIZE);
 FOFF_Est_Rx_Mat=repmat(FOFF_Est_Rx.',1,ML_FOFF1_BINS+1);
 FOFF_Est_Rx_Mat=FOFF_Est_Rx_Mat.*(FOFF_Est_Matrix.');
 FOFF_Est_Rx_Mat=[FOFF_Est_Rx_Mat;zeros(FOFF_EST_RX_SIZE,ML_FOFF1_BINS+1)];
//******************************************************************************
// ML frequency offset estimation, step 1. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6. To export the plot
// in various formats, give the following command in the scilab prompt:
// usecanvas(%F). For 'surf' plots, it is preferable to export in PNG format.
//******************************************************************************
 FOFF_Sync_Op=convol(MF_Coeff,FOFF_Est_Rx_Mat(:));
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op));
 Est_FOFF1_Inx=int((Max_Inx-1)/(2*FOFF_EST_RX_SIZE));
 Est_FOFF1=Est_FOFF1_Inx*Freq_Step-MAX_FOFF;
 Freq_Est_Err=Est_FOFF1-FOFF;
 RMS_Err_FOFF1=RMS_Err_FOFF1+Freq_Est_Err*Freq_Est_Err;
 Max_Err_FOFF1=max([Max_Err_FOFF1 abs(Freq_Est_Err)]);
//******************************************************************************
// Frame detection 1st attempt in the presence of frequency offset.
//******************************************************************************
 Diff_Inx=Max_Inx-PRE_AMB_SIZE-2*Est_FOFF1_Inx*FOFF_EST_RX_SIZE;
//******************************************************************************
// Channel estimation 1st attempt.
//******************************************************************************
 Demod_Rx=FOFF_Est_Matrix(Est_FOFF1_Inx+1,:).*FOFF_Est_Rx;
 if ((Diff_Inx >= 0) & (Diff_Inx < CHAN_LEN)) then
  Start_Inx1=CHAN_LEN+Diff_Inx;
  End_Inx1=Start_Inx1+PRE_AMB_SIZE-EST_CHAN_LEN;
  Rx_Sig1=Demod_Rx(Start_Inx1:End_Inx1);
  Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
//  Chan_Est_Ext_Zeros=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
//  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//  subplot(211);
//  plot(abs(Chan_FFT));
//  subplot(212);
//  plot(abs(Est_Chan_FFT))
  Rx_MF=convol(Chan_Est,Pre_Amb_IFFT);
  Rx_MF=conj([Rx_MF(RX_MF_SIZE:-1:1)]);         // Filter matched to rx. signal.
  Rx_MF_Energy=Rx_MF*Rx_MF';                    // Energy of the matched filter.
  Rx_MF=Rx_MF/Rx_MF_Energy;                        // Normalized to unit energy.
//******************************************************************************
  FOFF_Est_Rx_Mat2=repmat(Demod_Rx.',1,ML_FOFF2_BINS+1);
  FOFF_Est_Rx_Mat2=FOFF_Est_Rx_Mat2.*(FOFF_Est_Matrix2.');
  FOFF_Est_Rx_Mat2=[FOFF_Est_Rx_Mat2;zeros(FOFF_EST_RX_SIZE,ML_FOFF2_BINS+1)];
//******************************************************************************
//******************************************************************************
// ML frequency offset estimation, step 2. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6. To export the plot
// in various formats, give the following command in the scilab prompt:
// usecanvas(%F). For 'surf' plots, it is preferable to export in PNG format.
//******************************************************************************
  FOFF_Sync_Op2=convol(Rx_MF,FOFF_Est_Rx_Mat2(:));
  [Max_Sync_Op2,Max_Inx2]=max(abs(FOFF_Sync_Op2));
  [Max_Sync_Op3,Max_Inx3]=max(real(FOFF_Sync_Op2));
  Est_FOFF2_Inx=int((Max_Inx2-1)/(2*FOFF_EST_RX_SIZE));
  Est_FOFF3_Inx=int((Max_Inx3-1)/(2*FOFF_EST_RX_SIZE));
  Est_FOFF2=Est_FOFF2_Inx*Freq_Step2-MAX_FOFF2;
  Est_FOFF3=Est_FOFF3_Inx*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Est_FOFF1+Est_FOFF2-FOFF;
  Freq_Est_Err3=Est_FOFF1+Est_FOFF3-FOFF;
  RMS_Err_FOFF2=RMS_Err_FOFF2+Freq_Est_Err*Freq_Est_Err;
  RMS_Err_FOFF3=RMS_Err_FOFF3+Freq_Est_Err3*Freq_Est_Err3;
  Max_Err_FOFF2=max([Max_Err_FOFF2 abs(Freq_Est_Err)]);
  Max_Err_FOFF3=max([Max_Err_FOFF3 abs(Freq_Est_Err3)]);
//******************************************************************************
// Demodulation using the final estimate of the frequency offset.
//******************************************************************************
//  Est_FOFF1=0.0;
//  Est_FOFF2=0.0;
  Est_FOFF_Vec=0:FOFF_VEC_SIZE-1;
  Est_FOFF_Vec=Est_FOFF_Vec*%i*(Est_FOFF1+Est_FOFF2);
  Est_FOFF_Vec=exp(-Est_FOFF_Vec);
  Chan_Op=Chan_Op.*Est_FOFF_Vec;
//******************************************************************************
// Final channel estimation.
//******************************************************************************
  Rx_Sig1=Chan_Op(Start_Inx1:End_Inx1);
  Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
  Chan_Est_Ext_Zeros=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';             // Data size chan. FFT.
//******************************************************************************
// Plot the magnitude response of the actual and estimated channel.
//******************************************************************************
//  subplot(211);
//  plot(abs(Chan_FFT));
//  subplot(212);
//  plot(abs(Est_Chan_FFT))
//******************************************************************************
// Final data detection and interpolated FFT using zero-padding.
//******************************************************************************
  Start_Inx2=Start_Inx1+PRE_AMB_SIZE;
  End_Inx2=Start_Inx2+OFDM_SIZE-1;
  Rx_Sig2=Chan_Op(Start_Inx2:End_Inx2);                              // For FFT.
  Rx_Sig2=[Rx_Sig2 zeros(1,IP_FAC_ZP*OFDM_SIZE)];
  Rx_FFT_Op=fft(Rx_Sig2.',-1).';                     // FFT of received samples.
//******************************************************************************
// Get the noise free postamble part of the FFT output using the estimated
// channel and obtain the matched filter.
//******************************************************************************
  Post_FFT_Op=Post_Amb_Sym.*Est_Chan_FFT(Div_Post_Inx);
  OFDM_Post_FFT_Op=zeros(1,OFDM_SIZE);
  OFDM_Post_FFT_Op(BUF_SIZE+Div_Post_Inx)=Post_FFT_Op;
  Up_Post_FFT_Op=zeros(1,IP_FAC*OFDM_SIZE);
  Up_Post_FFT_Op(1:IP_FAC:$)=OFDM_Post_FFT_Op;
  Up_Post_FFT_Op_MF=conj(Up_Post_FFT_Op($:-1:1));
//******************************************************************************
// Get the resized matrix after FFT and deinterleaving.
//******************************************************************************
  Post_Corr_Op=convol(Up_Post_FFT_Op_MF,Rx_FFT_Op);
  [Post_Corr_Max,Post_Corr_Max_Inx]=max(abs(Post_Corr_Op));
  Diff_Inx1=Post_Corr_Max_Inx-IP_FAC*OFDM_SIZE;
//  pause;
//******************************************************************************
// Get the data part of the deinterleaved FFT output and the data part of the
// deinterleaved FFT of the channel estimate.
//******************************************************************************
  Rx_Data_Start=IP_FAC*BUF_SIZE+1+Diff_Inx1;
  Rx_Data_End=Rx_Data_Start+IP_FAC*PD_SIZE-1;
  Rx_FFT_Op1=Rx_FFT_Op(Rx_Data_Start:IP_FAC:Rx_Data_End);
  Div_Rx_FFT_Op=Rx_FFT_Op1(OFDM_Div_Map(1:DATA_SIZE));
  Div_Est_Chan_FFT=Est_Chan_FFT(BUF_SIZE+OFDM_Div_Map(1:DATA_SIZE));
//******************************************************************************
  Dist(1,:)=abs(Div_Rx_FFT_Op-QPSK_SYM(1,:).*Div_Est_Chan_FFT).^2;
  Dist(2,:)=abs(Div_Rx_FFT_Op-QPSK_SYM(2,:).*Div_Est_Chan_FFT).^2;
  Dist(3,:)=abs(Div_Rx_FFT_Op-QPSK_SYM(3,:).*Div_Est_Chan_FFT).^2;
  Dist(4,:)=abs(Div_Rx_FFT_Op-QPSK_SYM(4,:).*Div_Est_Chan_FFT).^2;
  [M,K]=min(Dist(1,:),Dist(2,:),Dist(3,:),Dist(4,:));  // Min. dist and indices.
  Rx_Sym=QPSK_Map(K(1,1:DATA_SIZE));                           // Column vector.
  Ber=Ber+nnz(real(Coded_Tx_Sym(1:DATA_SIZE)-Rx_Sym.'));
//******************************************************************************
// Compute noise variance in two dimensions, before the FFT operation.
//******************************************************************************
  N_Vec=Rx_Sig1.'-Chan_Ip_Matrix*Chan_Est;                      // Noise vector.
  Est_N_Var=N_Vec'*N_Vec/(PRE_AMB_SIZE-EST_CHAN_MEM);
//******************************************************************************
// Normalize the exponents. Collect the maximum from each column of My_Exp1
// and My_Exp2. Compute gammas.
//******************************************************************************
  My_Exp1_Data=zeros(CONSTELL_SIZE,DATA_SIZE);
  My_Exp2_Data=zeros(CONSTELL_SIZE,DATA_SIZE);
  My_Exp1= -Dist(:,1:DATA_SIZE/2)/(OFDM_SIZE*real(Est_N_Var));
  My_Exp2= -Dist(:,GINX1:GINX2)  /(OFDM_SIZE*real(Est_N_Var));
  My_Exp1=My_Exp1-repmat(max(My_Exp1,'r'),CONSTELL_SIZE,1);
  My_Exp2=My_Exp2-repmat(max(My_Exp2,'r'),CONSTELL_SIZE,1);
  Exp1_Inx=find(My_Exp1 < MIN_EXPO);
  Exp2_Inx=find(My_Exp2 < MIN_EXPO);
  My_Exp1(Exp1_Inx)= MIN_EXPO;
  My_Exp2(Exp2_Inx)= MIN_EXPO;
  My_Exp1_Data(:,1:2:$)=My_Exp1;
  My_Exp2_Data(:,1:2:$)=My_Exp2;
  My_Gamma1(:,:)=exp(My_Exp1_Data);
  My_Gamma2(:,:)=exp(My_Exp2_Data);
//  pause;
//******************************************************************************
// The MAP (BCJR) algorithm.
//******************************************************************************
  Prob=0.5*ones(2,DATA_SIZE);
//******************************************************************************
//  [Prob]=BCJR(Prob);
//******************************************************************************
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process_End(Prob);
//******************************************************************************
  [M1,K1]=max(Prob(1,:),Prob(2,:));                   // Max. prob. and indices.
  Rx_Sym1=1-2*(K1-1);
  C_Ber=C_Ber+nnz(My_Data-Rx_Sym1);
//******************************************************************************
 else
  Erasure=Erasure+1;
 end
end
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF2=sqrt(RMS_Err_FOFF2/(MAX_FRAMES-Erasure));
RMS_Err_FOFF3=sqrt(RMS_Err_FOFF3/(MAX_FRAMES-Erasure));
timer()
//******************************************************************************
