//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2011 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Frequency offset introduced 26/05/2011.
// Convolutional coding introduced 6/12/2011.
// Preamble amplitude adjustment (factor of 0.5) done on 02/04/2012.
// Code tested on SCILAB ver 5.3.3.
//******************************************************************************
clear;
timer();
getd(pwd());                      // Get all functions in the current directory.
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=1;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                     // Length of the preamble.
OFDM_SIZE=4096;           // Length of OFDM symbol, excluding the cyclic prefix.
PRE_AMB_QPSK_AMP=sqrt(0.5*PRE_AMB_SIZE/OFDM_SIZE); // Amp. of preamble QPSK sym.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=2*OFDM_SIZE+PRE_AMB_SIZE+EST_CHAN_MEM;                 // Frame size.
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF=0.04;           // Freq. offset lies in the interval [-0.04, 0.04] radians.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_LEN-1; // Size of rx. vec. for freq. est.
MAX_FOFF=0.04;                                            // Max . freq. offset.
MAX_FOFF2=0.005;                               // Max. freq. offset in 2nd step.
ML_FOFF1_BINS=64;                    // No. of freq. bins in 1st step of ML est.
ML_FOFF2_BINS=64;                    // No. of freq. bins in 2nd step of ML est.
if abs(FOFF) > MAX_FOFF then
 halt('Freq offset exceeds maximum.');
end;
//******************************************************************************
// SNR parameters.
//******************************************************************************
SNR_DB=0.0;                                                       // SNR in dB.
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));    // Var per dimension of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// Channel estimation. 0: estimated channel; 1: ideal channel.
//******************************************************************************
IDEAL_CHEST=0;
//******************************************************************************
// MMSE channel estimation. 0: least squares; 1: MMSE.
//******************************************************************************
MMSE_CHEST=0;
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
for i=1:CONSTELL_SIZE
 QPSK_SYM(i,:)=QPSK_Map(i)*ones(1,2*OFDM_SIZE);
end
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
[Re_Pre_Amb_Sym,Im_Pre_Amb_Sym,Pre_Amb_Sym]=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym=Pre_Amb_Sym*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1).'; // IFFT. There is division by PRE_AMB_SIZE.
//******************************************************************************
// Get channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE-EST_CHAN_MEM,EST_CHAN_LEN);     // Initialize.
for i=1:PRE_AMB_SIZE-EST_CHAN_MEM
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT(EST_CHAN_LEN+i-1:-1:i);
end
//******************************************************************************
// Least squares channel estimation matrix.
//******************************************************************************
if MMSE_CHEST == 0 then
 Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// MMSE channel estimation matrix.
//******************************************************************************
elseif MMSE_CHEST == 1 then
 R1=Chan_Ip_Matrix*Chan_Ip_Matrix';
 R=2*FADE_VAR*R1+2*NOISE_VAR*eye(R1);
 MMSE_Chan_Est_Matrix=Chan_Ip_Matrix'*inv(R);
end
//******************************************************************************
// Get the matched filter.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
MF_Coeff=MF_Coeff/MF_Energy;                       // Normalized to unit energy.
RX_MF_SIZE=FOFF_EST_RX_SIZE;                                  // Size of rx. MF.
//******************************************************************************
// Get the frequency offset matrix.
//******************************************************************************
FOFF_Vec=0:FOFF_VEC_SIZE-1;
FOFF_Vec=FOFF_Vec*%i*FOFF;
FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x;
//******************************************************************************
// Frequency offset estimation matrix for step 2.
//******************************************************************************
//MAX_FOFF2=2*Freq_Step;
Freq_Step2=2*MAX_FOFF2/ML_FOFF2_BINS;
y2=[0:ML_FOFF2_BINS]';
y2=y2*Freq_Step2-MAX_FOFF2;     // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix2=y2*x;
FOFF_Est_Matrix2=exp(-%i*FOFF_Est_Matrix2);
clear x;
//******************************************************************************
// Initialize frequency offset estimation parameters for step 1.
//******************************************************************************
PLOT_START_TIME=PRE_AMB_SIZE-EST_CHAN_LEN;
PLOT_END_TIME=PRE_AMB_SIZE+EST_CHAN_LEN;
PLOT_LEN=PLOT_END_TIME-PLOT_START_TIME+1;
//******************************************************************************
// Initialize frequency offset estimation parameters for step 2.
//******************************************************************************
PLOT_START_TIME2=RX_MF_SIZE-EST_CHAN_LEN;
PLOT_END_TIME2=RX_MF_SIZE+EST_CHAN_LEN;
PLOT_LEN2=PLOT_END_TIME2-PLOT_START_TIME2+1;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
RMS_Err_FOFF1=0.0;         // RMS frequency offset estimation error in 1st step.
Max_Err_FOFF1=-1.0;    // Maximum frequency offset estimation error in 1st step.
RMS_Err_FOFF2=0.0;         // RMS frequency offset estimation error in 2nd step.
Max_Err_FOFF2=-1.0;    // Maximum frequency offset estimation error in 2nd step.
//******************************************************************************
// Arrays for the 3D plot of the frequency offset estimate. 'y', 'y2' and
// 'y3' arrays already defined above.
//******************************************************************************
x=[PLOT_START_TIME:PLOT_END_TIME];                           // x range, step 1.
x2=[PLOT_START_TIME2:PLOT_END_TIME2];                        // x range, step 2.
z=zeros(ML_FOFF1_BINS+1,PLOT_LEN);                           // 3D plot, step 1.
z2=zeros(ML_FOFF2_BINS+1,PLOT_LEN2);                         // 3D plot, step 2.
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D^2+D^3+D^4)/(1+D+D^4)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 0 1];                      // Numerator coefficients in reverse order.
DC=[1 1 1];                    // Denominator coefficients in reverse order.
NP=poly(NC,'D','coeff');                                // Numerator polynomial.
DP=poly(DC,'D','coeff');                              // Denominator polynomial.
RSC_Gen_Vec=modulo(abs(ldiv(NP,DP,OFDM_SIZE)),2);     // Column vector of coeff.
NC_Size=size(NC);
NUM_INPUTS=2;                                      // No. of input combinations.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op]=Get_Trellis(NC,DC,NUM_INPUTS,NUM_STATES,RSC_MEM);
//******************************************************************************
// Get the interleaver and deinterleaver map.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(OFDM_SIZE);
//******************************************************************************
// Cyclic prefix (CP) indices.
//******************************************************************************
CP_ST_INX=2*OFDM_SIZE-EST_CHAN_MEM+1;       // Start index for extraction of CP.
CP_ED_INX=2*OFDM_SIZE;                        // End index for extraction of CP.
//******************************************************************************
// Number of iterations of the BCJR algorithm.
//******************************************************************************
NUM_ITER=8;
//******************************************************************************
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Erasure=0;                                          // Number of frame erasures.
Ber_Vec=zeros(1,MAX_FRAMES);
C_Ber_Vec=zeros(1,MAX_FRAMES);
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
// clf();
//******************************************************************************
// Get the random frequency offset and the frequency offset matrix.
//******************************************************************************
 FOFF=2.0*MAX_FOFF*(grand(1,1,'def')-0.5);
 if abs(FOFF) > MAX_FOFF then
  halt('Freq offset exceeds maximum.');
 end;
 FOFF_Vec=0:FOFF_VEC_SIZE-1;
 FOFF_Vec=FOFF_Vec*%i*FOFF;
 FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Initialize estimates of the frequency offset.
//******************************************************************************
 Est_FOFF1=0.0;
 Est_FOFF2=0.0;
//******************************************************************************
// Generate coded QPSK symbols.
//******************************************************************************
 //[Re_Tx_Sym,Im_Tx_Sym,Tx_Sym]=Get_QPSK_Sym(OFDM_SIZE);
 [Tx_Sym]=Get_Coded_QPSK(OFDM_SIZE,RSC_Gen_Vec,Div_Map);
//******************************************************************************
 Sym_IFFT=fft(Tx_Sym.',1).';            // IFFT. There is division by OFDM_SIZE.
//******************************************************************************
// Add cyclic prefix and preamble.
//******************************************************************************
 Chan_Ip=[Pre_Amb_IFFT Sym_IFFT(CP_ST_INX:CP_ED_INX) Sym_IFFT];
//******************************************************************************
// Get channel.
//******************************************************************************
 [Re_Chan,Im_Chan,Chan]=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan zeros(1,2*OFDM_SIZE-CHAN_LEN)];           // Zero padding.
 Chan_FFT=fft(Chan_Ext_Zeros.',-1).';                  // 2*OFDM_SIZE-point FFT.
//******************************************************************************
 Chan_Op=Get_Channel_Op(Chan,Chan_Ip,FOFF_Vec,NOISE_MEAN,NOISE_STD);
 FOFF_Est_Rx=[Chan_Op(1:FOFF_EST_RX_SIZE)];
//******************************************************************************
// ML frequency offset estimation, step 1. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6. To export the plot
// in various formats, give the following command in the scilab prompt:
// usecanvas(%F). For 'surf' plots, it is preferable to export in PNG format.
//******************************************************************************
 for k=1:ML_FOFF1_BINS+1
//  Demod_Rx=FOFF_Est_Matrix(k,:).*FOFF_Est_Rx;
  FOFF_Sync_Op=convol(MF_Coeff,FOFF_Est_Matrix(k,:).*FOFF_Est_Rx);
  z(k,:)=FOFF_Sync_Op(PLOT_START_TIME:PLOT_END_TIME);
 end;
 [Max_Sync_Op,Max_Inx]=max(abs(z));
 Est_FOFF1=(Max_Inx(1)-1)*Freq_Step-MAX_FOFF;
 Freq_Est_Err=Est_FOFF1-FOFF;
 RMS_Err_FOFF1=RMS_Err_FOFF1+Freq_Est_Err*Freq_Est_Err;
 Max_Err_FOFF1=max([Max_Err_FOFF1 abs(Freq_Est_Err)]);
 clear FOFF_Sync_Op;
 subplot(211);
 surf(x,y',abs(z).^2,'linestyle','dot','edgecol','red','facecol','inter');
// plot3d(y,log10(x),abs(z).^2,flag=[0,4,3]);
// plot3d1(y,log10(x),abs(z).^2);
//******************************************************************************
// Frame detection 1st attempt in the presence of frequency offset.
//******************************************************************************
 if IDEAL_CHEST == 0 then
  Diff_Inx=Max_Inx(2)-EST_CHAN_LEN-1;
 else
  Diff_Inx=CHAN_MEM;
 end
//******************************************************************************
// Channel estimation 1st attempt.
//******************************************************************************
 Demod_Rx=FOFF_Est_Matrix(Max_Inx(1),:).*FOFF_Est_Rx;
 if ((Diff_Inx >= 0) & (Diff_Inx < CHAN_LEN)) then
  Start_Inx1=CHAN_LEN+Diff_Inx;
  End_Inx1=Start_Inx1+PRE_AMB_SIZE-EST_CHAN_LEN;
  Rx_Sig1=Demod_Rx(Start_Inx1:End_Inx1);
  if MMSE_CHEST == 0 then
   Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
  elseif MMSE_CHEST == 1 then
   Chan_Est=MMSE_Chan_Est_Matrix*Rx_Sig1.';
  end
//  Chan_Est_Ext_Zeros=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
//  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//  subplot(211);
//  plot(abs(Chan_FFT));
//  subplot(212);
//  plot(abs(Est_Chan_FFT))
  Rx_MF=convol(Chan_Est,Pre_Amb_IFFT);
  Rx_MF=conj([Rx_MF(RX_MF_SIZE:-1:1)]);         // Filter matched to rx. signal.
  Rx_MF_Energy=Rx_MF*Rx_MF';                    // Energy of the matched filter.
  Rx_MF=Rx_MF/Rx_MF_Energy;                        // Normalized to unit energy.
//******************************************************************************
// ML frequency offset estimation, step 2. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6. To export the plot
// in various formats, give the following command in the scilab prompt:
// usecanvas(%F). For 'surf' plots, it is preferable to export in PNG format.
//******************************************************************************
  for k=1:ML_FOFF2_BINS+1
//   Demod_Rx2=FOFF_Est_Matrix2(k,:).*Demod_Rx;
   FOFF_Sync_Op=convol(Rx_MF,FOFF_Est_Matrix2(k,:).*Demod_Rx);
   z2(k,:)=FOFF_Sync_Op(PLOT_START_TIME2:PLOT_END_TIME2);
  end;
  [Max_Sync_Op2,Max_Inx2]=max(abs(z2));
  Est_FOFF2=(Max_Inx2(1)-1)*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Est_FOFF1+Est_FOFF2-FOFF;
  RMS_Err_FOFF2=RMS_Err_FOFF2+Freq_Est_Err*Freq_Est_Err;
  Max_Err_FOFF2=max([Max_Err_FOFF2 abs(Freq_Est_Err)]);
  clear FOFF_Sync_Op;
  subplot(212);
  surf(x2,y2',abs(z2).^2,'linestyle','dot','edgecol','red','facecol','inter');
//  plot3d(y,log10(x),abs(z2).^2,flag=[0,4,3]);
//  plot3d1(y,log10(x),abs(z2).^2);
//******************************************************************************
// Demodulation using the final estimate of the frequency offset.
//******************************************************************************
  Est_FOFF_Vec=0:FOFF_VEC_SIZE-1;
  Est_FOFF_Vec=Est_FOFF_Vec*%i*(Est_FOFF1+Est_FOFF2);
  Est_FOFF_Vec=exp(-Est_FOFF_Vec);
  Chan_Op=Chan_Op.*Est_FOFF_Vec;
//******************************************************************************
// Final channel estimation.
//******************************************************************************
  Rx_Sig1=Chan_Op(Start_Inx1:End_Inx1);
  if MMSE_CHEST == 0 then
   Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
  elseif MMSE_CHEST == 1 then
   Chan_Est=MMSE_Chan_Est_Matrix*Rx_Sig1.';
  end
  Chan_Est_Ext_Zeros=[Chan_Est;zeros(2*OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';             // Data size chan. FFT.
//******************************************************************************
// Plot the magnitude response of the actual and estimated channel.
//******************************************************************************
//  subplot(211);
//  plot(abs(Chan_FFT));
//  subplot(212);
//  plot(abs(Est_Chan_FFT))
//******************************************************************************
// Final data detection.
//******************************************************************************
  Start_Inx2=Start_Inx1+PRE_AMB_SIZE;
  End_Inx2=Start_Inx2+2*OFDM_SIZE-1;
  Rx_Sig2=Chan_Op(Start_Inx2:End_Inx2);                              // For FFT.
  Rx_FFT_Op=fft(Rx_Sig2.',-1).';                     // FFT of received samples.
  if IDEAL_CHEST == 0 then                        // Compute sq. Euclidean dist.
   for i=1:CONSTELL_SIZE
    Dist(i,:)=abs(Rx_FFT_Op-QPSK_SYM(i,:).*Est_Chan_FFT).^2;
   end
  else
   for i=1:CONSTELL_SIZE
    Dist(i,:)=abs(Rx_FFT_Op-QPSK_SYM(i,:).*Chan_FFT).^2;
   end
  end
  [M,K]=min(Dist(1,:),Dist(2,:),Dist(3,:),Dist(4,:));  // Min. dist and indices.
  for i=1:OFDM_SIZE                              // Map indices to QPSK symbols.
   Rx_Sym(1,i)=QPSK_Map(K(1,i));                                  // Row vector.
  end
  F_Ber=nnz(real(Tx_Sym(1:OFDM_SIZE)-Rx_Sym));
  Ber=Ber+F_Ber;
  Ber_Vec(1,Frame_Cnt)=F_Ber;
//******************************************************************************
// Compute noise variance in two dimensions, before the FFT operation.
//******************************************************************************
  N_Vec=Rx_Sig1.'-Chan_Ip_Matrix*Chan_Est;                      // Noise vector.
  Est_N_Var=N_Vec'*N_Vec/(PRE_AMB_SIZE-EST_CHAN_MEM);
//******************************************************************************
// Compute gammas.
//******************************************************************************
  ginx1=OFDM_SIZE+1;
  ginx2=2*OFDM_SIZE;
  for i=1:4
   My_Gamma1(i,:)=exp(-Dist(i,1:OFDM_SIZE)/(2*OFDM_SIZE*real(Est_N_Var)));
   My_Gamma2(i,:)=exp(-Dist(i,ginx1:ginx2)/(2*OFDM_SIZE*real(Est_N_Var)));
  end
//******************************************************************************
// The MAP (BCJR) algorithm.
//******************************************************************************
  Prob=0.5*ones(2,OFDM_SIZE);
  [Prob]=BCJR(Prob);
  [M1,K1]=max(Prob(1,:),Prob(2,:));                   // Max. prob. and indices.
  Rx_Sym1=1-2*(K1-1);
  FC_Ber=nnz(real(Tx_Sym(1:OFDM_SIZE)-Rx_Sym1));
  C_Ber=C_Ber+FC_Ber;
  C_Ber_Vec(1,Frame_Cnt)=FC_Ber;
  if Frame_Cnt == 152
   pause;
  end
//******************************************************************************
 else
  Erasure=Erasure+1;
 end
end
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF2=sqrt(RMS_Err_FOFF2/(MAX_FRAMES-Erasure));
fd1=mopen('ber_vec.dat','w');
fd2=mopen('c_ber_vec.dat','w');
for i=1:MAX_FRAMES
 mfprintf(fd1,'%d %d\n',i,Ber_Vec(1,i));
 mfprintf(fd2,'%d %d\n',i,C_Ber_Vec(1,i));
end
timer()
//******************************************************************************
